from abc import ABCMeta, abstractmethod
from typing import Dict


class Repository:

    __metaclass__ = ABCMeta

    @abstractmethod
    def info(self) -> Dict:
        return {"Repository": "Not Implemented"}



import connexion

from flask_cors import CORS

from petisco.frameworks.flask.application.json_encoder import JSONEncoder


class FlaskApplication:
    def __init__(self, application_name: str, swagger_dir: str, port: int = 8080):
        self.application_name = application_name
        self.app = connexion.App(__name__, specification_dir=swagger_dir)
        self.port = port
        self._configure()

    def _configure(self):
        CORS(self.app.app)
        self.app.app.json_encoder = JSONEncoder
        self.app.add_api("swagger.yaml", arguments={"title": self.application_name})

    def start(self):
        self.app.run(port=self.port)

    def get_config(self):
        return self.config

    def get_app(self):
        return self.app

    config = property(get_config)



import threading
from typing import Dict, Callable

from pika import BlockingConnection, ConnectionParameters, BasicProperties

from petisco.events.event import Event
from petisco.events.interface_event_manager import IEventManager


class RabbitMQEventManager(IEventManager):
    def __init__(
        self,
        connection_parameters: ConnectionParameters,
        subscribers: Dict[str, Callable] = None,
    ):
        super().__init__(subscribers)
        self.connection_parameters = connection_parameters

        if self.subscribers:
            # Run the worker into a thread
            self._thread = threading.Thread(target=self._subscribe)
            self._thread.start()

    def info(self) -> Dict:
        return {
            "name": self.__class__.__name__,
            "host": self.connection_parameters._host,
            "port": self.connection_parameters._port,
        }

    def _subscribe(self):
        self._connection_subscriber = BlockingConnection(self.connection_parameters)
        self._channel_subscriber = self._connection_subscriber.channel()
        self._channel_subscriber.basic_qos(prefetch_count=1)

        for topic, callback in self.subscribers.items():
            self._channel_subscriber.queue_declare(queue=topic, durable=True)
            self._channel_subscriber.basic_consume(
                queue=topic, on_message_callback=callback
            )

        self._channel_subscriber.start_consuming()

    def unsubscribe_all(self):
        def kill():
            self._channel_subscriber.stop_consuming()

        if self.subscribers:
            self._connection_subscriber.call_later(0, kill)
            self._thread.join()

    def send(self, topic: str, event: Event):
        connection_publisher = BlockingConnection(self.connection_parameters)
        channel = connection_publisher.channel()
        channel.queue_declare(queue=topic, durable=True)
        channel.basic_publish(
            exchange="",
            routing_key=topic,
            body=event.to_json(),
            properties=BasicProperties(delivery_mode=2),  # make message persistent
        )
        connection_publisher.close()



import base64
import json


def api_key_info_func(api_key, required_scopes):
    """
    Check and retrieve authentication information from api_key.
    Returned value will be passed in 'token_info' parameter of your operation function, if there is one.
    'sub' or 'uid' will be set in 'user' parameter of your operation function, if there is one.

    :param api_key API key provided by Authorization header
    :type api_key: str
    :param required_scopes Always None. Used for other authentication method
    :type required_scopes: None
    :return: Information attached to provided api_key or None if api_key is invalid or does not allow access to called API
    :rtype: dict | None
    """
    return {"uid": api_key}


def bearer_info_func(token):
    """
    Check and retrieve authentication information from custom bearer token.
    Returned value will be passed in 'token_info' parameter of your operation function, if there is one.
    'sub' or 'uid' will be set in 'user' parameter of your operation function, if there is one.

    :param token Token provided by Authorization header
    :type token: str
    :return: Decoded token information or None if token is invalid
    :rtype: dict | None
    """
    token_payload = decode_token_payload(token)
    token_info = {
        "user_id": token_payload.get("sub"),
        "client_id": token_payload.get("cli"),
        "token_type": token_payload.get("typ"),
    }
    return token_info


def decode_token_payload(token):
    try:
        token_payload = token.split(".")[1]
    except IndexError:
        return None
    token_payload += "=" * ((4 - len(token_payload) % 4) % 4)
    return json.loads(base64.b64decode(token_payload).decode("utf-8"))



from typing import Dict, Callable
from redis import Redis

from petisco.events.event import Event
from petisco.events.interface_event_manager import IEventManager


class RedisEventManager(IEventManager):
    def __init__(self, redis: Redis, subscribers: Dict[str, Callable] = None):
        super().__init__(subscribers)
        self._redis = redis

        if self.subscribers:
            self._pubsub = self._redis.pubsub()
            self._subscribe()

    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def _subscribe(self):
        self._pubsub.subscribe(**self.subscribers)
        self._thread = self._pubsub.run_in_thread(sleep_time=0.001)

    def unsubscribe_all(self):
        if self.subscribers:
            self._thread.stop()

    def send(self, topic: str, event: Event):
        self._redis.publish(topic, event.to_json())



import inspect
from functools import wraps
import traceback
from typing import Callable, Tuple, Dict, List, Any

from meiga import Result
from meiga.decorators import meiga

from petisco.controller.errors.internal_http_error import InternalHttpError
from petisco.controller.errors.known_result_failure_handler import (
    KnownResultFailureHandler,
)
from petisco.controller.tokens.jwt_decorator import jwt
from petisco.logger.interface_logger import ERROR, INFO
from petisco.controller.errors.http_error import HttpError
from petisco.controller.tokens.jwt_config import JwtConfig
from petisco.frameworks.flask.correlation_id_provider import (
    flask_correlation_id_provider,
)
from petisco.logger.log_message import LogMessage
from petisco.logger.not_implemented_logger import NotImplementedLogger

DEFAULT_SUCCESS_MESSAGE = {"message": "OK"}, 200
DEFAULT_ERROR_MESSAGE = HttpError().handle()


class _ControllerHandler:
    def __init__(
        self,
        logger=NotImplementedLogger(),
        jwt_config: JwtConfig = None,
        success_handler: Callable[[Result], Tuple[Dict, int]] = None,
        error_handler: Callable[[Result], HttpError] = None,
        correlation_id_provider: Callable = flask_correlation_id_provider,
        logging_types_blacklist: List[Any] = [bytes],
    ):
        self.logger = logger
        self.jwt_config = jwt_config
        self.success_handler = success_handler
        self.error_handler = error_handler
        self.correlation_id_provider = correlation_id_provider
        self.logging_types_blacklist = logging_types_blacklist

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):
            @meiga
            @jwt(jwt_config=self.jwt_config)
            def run_controller(*args, **kwargs) -> Result:
                return func(*args, **kwargs)

            correlation_id, kwargs = update_correlation_id(kwargs)
            log_message = LogMessage(
                layer="controller",
                operation=f"{func.__name__}",
                correlation_id=correlation_id,
            )

            try:
                log_message.message = "Start"
                self.logger.log(INFO, log_message.to_json())
                result = run_controller(*args, **kwargs)
                if result.is_success:
                    if isinstance(result.value, tuple(self.logging_types_blacklist)):
                        log_message.message = (
                            f"Success result of type: {type(result.value).__name__}"
                        )
                    else:
                        log_message.message = f"{result}"
                    self.logger.log(INFO, log_message.to_json())
                    return (
                        self.success_handler(result)
                        if self.success_handler
                        else DEFAULT_SUCCESS_MESSAGE
                    )
                else:
                    log_message.message = f"{result}"
                    self.logger.log(ERROR, log_message.to_json())
                    known_result_failure_handler = KnownResultFailureHandler(result)

                    if not known_result_failure_handler.is_a_result_known_error:
                        if not self.error_handler:
                            return DEFAULT_ERROR_MESSAGE

                        http_error = self.error_handler(result)

                        if not issubclass(http_error.__class__, HttpError):
                            raise TypeError(
                                "Returned object from error_handler must be subclasses of HttpError"
                            )

                        return http_error.handle()
                    else:
                        return known_result_failure_handler.http_error.handle()

            except Exception as e:
                log_message.message = f"Error {func.__name__}: {repr(e.__class__)} {e} | {traceback.format_exc()}"
                self.logger.log(ERROR, log_message.to_json())
                return InternalHttpError().handle()

        def update_correlation_id(kwargs):
            signature = inspect.signature(func)
            if (
                "correlation_id" in signature.parameters
                and "correlation_id" not in kwargs
            ):
                correlation_id = self.correlation_id_provider(func.__name__)
                kwargs = dict(kwargs, correlation_id=correlation_id)
            else:
                correlation_id = None
            return correlation_id, kwargs

        return wrapper


controller_handler = _ControllerHandler



from typing import Dict, Callable
from abc import ABCMeta, abstractmethod

from petisco.events.event import Event


class IEventManager:

    __metaclass__ = ABCMeta

    def __init__(self, subscribers: Dict[str, Callable]):
        self.subscribers = subscribers

    def __repr__(self):
        return f"EventManager: [subscribers: {self.subscribers}]"

    @abstractmethod
    def info(self) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_all(self):
        raise NotImplementedError

    @abstractmethod
    def send(self, topic: str, event: Event):
        raise NotImplementedError



from typing import Callable

from petisco.persistence.interface_persistence_connector import IPersistenceConnector
from petisco.persistence.sqlalchemy.sqlalchemy_persistence import SqlAlchemyPersistence
from petisco.persistence.sqlalchemy.sqlalchemy_persistence_config import (
    SqlAlchemyPersistenceConfig,
)


class SqlAlchemyPersistenceConnector(IPersistenceConnector):
    def __init__(
        self,
        config: SqlAlchemyPersistenceConfig,
        import_database_models: Callable = None,
    ):
        self.config = config
        self.import_database_models = import_database_models

    def get_connection(self):
        connection = None
        if self.config.server:
            if self.config.server == "sqlite":
                connection = "{}:///{}".format(self.config.server, self.config.database)
            elif self.config.server == "mysql":
                connection = "{}+{}://{}:{}@{}:{}/{}".format(
                    self.config.server,
                    self.config.driver,
                    self.config.user,
                    self.config.password,
                    self.config.host,
                    self.config.port,
                    self.config.database,
                )
        return connection

    def execute(self):
        from sqlalchemy import create_engine
        from sqlalchemy.ext.declarative import declarative_base
        from sqlalchemy.orm import sessionmaker
        from sqlalchemy_utils import database_exists, create_database

        connection = self.get_connection()

        persistence = SqlAlchemyPersistence(base=declarative_base())

        self.import_database_models()

        if connection:
            if self.config.server == "sqlite":
                engine = create_engine(
                    connection,
                    json_serializer=lambda obj: obj,
                    json_deserializer=lambda obj: obj,
                )
            else:
                engine = create_engine(
                    connection,
                    pool_pre_ping=True,
                    json_serializer=lambda obj: obj,
                    json_deserializer=lambda obj: obj,
                )

            if not database_exists(engine.url):
                create_database(engine.url)
                persistence.base.metadata.create_all(engine)

            persistence.session = sessionmaker(bind=engine)



from datetime import datetime
from typing import Dict

from petisco.controller.correlation_id import CorrelationId
from petisco.events.event_id import EventId

import json


class Event:
    id: EventId = None
    name: str = None
    occurred_on: str = None
    version: str = None
    correlation_id: CorrelationId = None

    def __init__(self, dictionary=None):
        if dictionary:
            self.__dict__.update(dictionary)

        self.id = EventId.generate(str(dictionary)) if not self.id else self.id
        self.name = self.__class__.__name__ if not self.name else self.name
        self.occurred_on = (
            datetime.utcnow() if not self.occurred_on else self.occurred_on
        )

    def __repr__(self):
        return f"[{self.name}: {json.loads(self.to_json())}]"

    def __eq__(self, other):
        if (
            isinstance(other, self.__class__)
            or issubclass(other.__class__, self.__class__)
            or issubclass(self.__class__, other.__class__)
        ):
            return self.to_dict() == other.to_dict()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_dict(dictionary: Dict):
        if "id" in dictionary and isinstance(dictionary["id"], str):
            dictionary["id"] = EventId(dictionary["id"])
        if "occurred_on" in dictionary and isinstance(dictionary["occurred_on"], str):
            dictionary["occurred_on"] = datetime.strptime(
                dictionary["occurred_on"], "%Y-%m-%d %H:%M:%S.%f"
            )
        return Event(dictionary)

    def to_json(self):
        return json.dumps(self.to_dict(), default=self._datetime_to_str)

    def _datetime_to_str(self, o):
        if isinstance(o, datetime):
            return o.__str__()

    @staticmethod
    def from_json(event_json):
        event_dict = json.loads(event_json)
        return Event.from_dict(event_dict)



import hashlib
import random
import string
from datetime import datetime


class CorrelationId(str):
    def __init__(self, value):
        super(CorrelationId, self).__init__()
        self.value = value

    @staticmethod
    def generate(func_name: str = ""):
        datetime_string = datetime.utcnow().strftime("%d/%m/%Y %H:%M:%S")
        data_to_hash = (
            func_name + CorrelationId.random_string(10) + " " + datetime_string
        )
        return CorrelationId(CorrelationId.hash_string(data_to_hash))

    @staticmethod
    def hash_string(string: str):
        """
        Return a SHA-256 hash of the given string
        """
        return hashlib.sha256(string.encode("utf-8")).hexdigest()[:5]

    @staticmethod
    def random_string(string_length: int):
        letters = string.ascii_letters
        return "".join(random.choice(letters) for i in range(string_length))
